module SPLib.List 
  (isEmpty, length, get, getD, getP, first, firstD, firstP, SPLib.List.last, lastD, lastP, set, setD, setP, prepend, append, insert, insertD, insertP, concat, remove, removeD, removeP, removeFirst, removeFirstD, removeFirstP, removeLast, removeLastD, removeLastP, contains, removeDuplicates, L.sort, L.sortBy, reverse, take, drop, splitAt, zip, map, filter, reduce, reduceD, reduceP, foldl, foldr, SPLib.List.find, findD, findP, SPLib.List.findIndex, findIndexD, findIndexP, repeat, join
  ) where

import Prelude as P (elem, reverse, take, drop, zip, map, filter, foldl, foldr, null, length, head, last, concat, tail, init, splitAt, repeat, (!!)) 
import qualified Data.List as L (sort, sortBy, find, findIndex, intercalate, nub) 

import SPLib.Basic
import SPLib.PartialFunction

type CallLocation = String

--SIGNATURES

isEmpty   :: [a] -> Bool
--length  :: [a] -> Int 

get      :: Int -> [a] -> a  
getD     :: CallLocation -> Int -> [a] -> a  
getP     :: PartialFunction (Int, [a]) a   
first    :: [a] -> a 
firstD   :: CallLocation -> [a] -> a  
firstP   :: PartialFunction [a] a
--last   :: [a] -> a
lastD    :: CallLocation -> [a] -> a
lastP    :: PartialFunction [a] a 

set       :: a -> Int -> [a] -> [a]
setD      :: CallLocation -> a -> Int -> [a] -> [a]
setP      :: PartialFunction (a, Int, [a]) [a]
prepend   :: a -> [a] -> [a]
append    :: a -> [a] -> [a]
insert    :: a -> Int -> [a] -> [a]  -- (get i (insert x i list)) = x
insertD   :: CallLocation -> a -> Int -> [a] -> [a] 
insertP   :: PartialFunction (a, Int, [a]) [a]
--concat  :: [[a]] -> [a]

remove       :: Int -> [a] -> [a]        
removeD      :: CallLocation -> Int -> [a] -> [a]        
removeP      :: PartialFunction (Int, [a]) [a]        
removeFirst  :: [a] -> [a]
removeFirstD :: CallLocation -> [a] -> [a]
removeFirstP :: PartialFunction [a] [a]
removeLast   :: [a] -> [a]    
removeLastD  :: CallLocation -> [a] -> [a]    
removeLastP  :: PartialFunction [a] [a]    
        
contains  :: Eq a => a -> [a] -> Bool
removeDuplicates :: Eq a => [a] -> [a] 
--sort    :: Ord a => [a] -> [a]
--reverse :: [a] -> [a]    
--take    :: Int -> [a] -> [a]
--drop    :: Int -> [a] -> [a]
--splitAt :: Int -> [a] -> ([a],[a])
--zip     :: [a] -> [b] -> [(a,b)]   
    
--map     :: (a -> b) -> [a] -> [b]
--filter  :: (a -> Bool) -> [a] -> [a] 
reduce    :: (a -> a -> a) -> [a] -> a   
reduceD   :: CallLocation -> (a -> a -> a) -> [a] -> a   
reduceP   :: PartialFunction ((a -> a -> a),[a]) a
--foldl   :: (b -> a -> b) -> b -> [a] -> b
--foldr   :: (a -> b -> b) -> b -> [a] -> b

find       :: (a -> Bool) -> [a] -> a 
findD      :: CallLocation -> (a -> Bool) -> [a] -> a 
findP      :: PartialFunction ((a -> Bool),[a]) a
findIndex  :: (a -> Bool) -> [a] -> Int
findIndexD :: CallLocation -> (a -> Bool) -> [a] -> Int
findIndexP :: PartialFunction ((a -> Bool),[a]) Int

--repeat :: a -> [a]

join          :: [a] -> [[a]] -> [a]           


-- DEFINITIONS 

isEmpty = P.null
--length = P.length    
    
get  = unpack2 . apply $ getP
getD cl = unpack2 . (apply' ("SPLib.List.getD @ "++cl)) $ getP
getP = PartialFunction "SPLib.List.getP" f f_ood f_err
  where
    f (index,list) = list!!index
    f_ood (index,list) = P.concat [r1,r2]
      where
        r1 = if' (index < 0)              [(0,[(show index),(show (P.length list))])] []
        r2 = if' (index >= (P.length list)) [(1,[(show index),(show (P.length list))])] []
    f_err = 
      [
        ["negative index ","; list has "," elements"],
        ["index "," out of bounds; list has "," elements"]
      ]

first = apply firstP 
firstD cl = apply' ("SPLib.List.firstD @ "++cl) firstP
firstP = PartialFunction "SPLib.List.firstP" f f_ood f_err
  where
    f list = head list
    f_ood list = if' (isEmpty list) [(0,[])] []
    f_err = [["empty list"]]

last = apply lastP 
lastD cl = apply' ("SPLib.List.lastD @ "++cl) lastP
lastP = PartialFunction "SPLib.List.lastP" f f_ood f_err
  where
    f list = P.last list
    f_ood list = if' (isEmpty list) [(0,[])] []
    f_err = [["empty list"]]


set  = unpack3 . apply $ setP
setD cl = unpack3 . (apply' ("SPLib.List.setD @ "++cl)) $ setP
setP = PartialFunction "SPLib.List.setP" f f_ood f_err
  where
    f (val,index,list) = x++(prepend val (removeFirst y))
      where
        (x,y) = (P.splitAt index list)
    f_ood (_,index,list) = P.concat [r1,r2]
      where
        r1 = if' (index < 0)              [(0,[(show index),(show (P.length list))])] []
        r2 = if' (index >= (P.length list)) [(1,[(show index),(show (P.length list))])] []
    f_err = 
      [
        ["negative index ","; list has "," elements"],
        ["index "," out of bounds; list has "," elements"]
      ]
      
prepend val list = val:list
append  val list = list++[val]
insert  = unpack3 . apply $ insertP
insertD cl = unpack3 . (apply' ("SPLib.List.insertD @ "++cl)) $ insertP
insertP = PartialFunction "SPLib.List.insertP" f f_ood f_err
  where
    f (val,index,list) = x++(prepend val y)
      where
        (x,y) = (P.splitAt index list)
    f_ood (_,index,list) = P.concat [r1,r2]
      where
        r1 = if' (index < 0)              [(0,[(show index),(show (P.length list))])] []
        r2 = if' (index > (P.length list)) [(1,[(show index),(show (P.length list))])] []
    f_err = 
      [
        ["negative index ","; list has "," elements"],
        ["index "," out of bounds; list has "," elements (that is also the largest legal index)"]
      ]    
    
remove = unpack2 . apply $ removeP
removeD cl = unpack2 . (apply' ("SPLib.List.removeD @ "++cl)) $ removeP
removeP = PartialFunction "SPLib.List.removeP" f f_ood f_err
  where
    f (index,list) = x++(removeFirst y)
      where
        (x,y) = (P.splitAt index list) 
    f_ood (index,list) = P.concat [r1,r2]
      where
        r1 = if' (index < 0)              [(0,[(show index),(show (P.length list))])] []
        r2 = if' (index >= (P.length list)) [(1,[(show index),(show (P.length list))])] []
    f_err = 
      [
        ["negative index ","; list has "," elements"],
        ["index "," out of bounds; list has "," elements"]
      ]    
    
removeFirst = apply removeFirstP
removeFirstD cl = (apply' ("SPLib.List.removeFirstD @ "++cl)) removeFirstP
removeFirstP = PartialFunction "SPLib.List.removeFirstP" f f_ood f_err
  where
    f list = P.tail list
    f_ood list = if' (isEmpty list) [(0,[])] []
    f_err = [["empty list"]]
    
removeLast  = apply removeLastP
removeLastD cl = (apply' ("SPLib.List.removeLastD @ "++cl)) removeLastP
removeLastP = PartialFunction "SPLib.List.removeLastP" f f_ood f_err
  where
    f list = P.init list
    f_ood list = if' (isEmpty list) [(0,[])] []
    f_err = [["empty list"]]
        
contains  = P.elem
removeDuplicates = L.nub

reduce = unpack2 . apply $ reduceP
reduceD cl = unpack2 . (apply' ("SPLib.List.reduceD @ "++cl)) $ reduceP
reduceP = (PartialFunction "SPLib.List.reduceP" f f_ood f_err)
  where
    f (f,list) = P.foldl f (first list) (removeFirst list) 
    f_ood (_,list) = if' (isEmpty list) [(0,[])] []
    f_err = [["can't reduce empty list"]]

find      = unpack2 . apply $ findP
findD cl  = unpack2 . (apply' ("SPLib.List.findD @ "++cl)) $ findP
findIndex = unpack2 . apply $ findIndexP
findIndexD cl = unpack2 . (apply' ("SPLib.List.findIndexD @ "++cl)) $ findIndexP
findP = (PartialFunction "SPLib.List.findP" f f_ood f_err)
  where
    f (pred,list) = (just (L.find pred list))
    f_ood (pred,list) = if' (isNothing (L.find pred list)) [(0,[])] []
    f_err = [["no element found which matchtes the predicate"]]
findIndexP = (PartialFunction "SPLib.List.findIndexP" f f_ood f_err)
  where
    f (pred,list) = (just (L.findIndex pred list))
    f_ood (pred,list) = if' (isNothing (L.findIndex pred list)) [(0,[])] []
    f_err = [["no element found which matches the predicate"]]
    
join  = L.intercalate
